#include <glib-unix.h>
#include "shell.h"
#include "wayland.h"
#include "virtual-keyboard.h"
enum
{
    NO_PROP,
    PROP_SHELL_WAYLAND,
    PROP_SHELL_APP,
    PROP_SHELL_NAME,
    N_PROPERTIES
};

static GParamSpec *shell_properties[N_PROPERTIES] = {0};


typedef struct _DiyaShellPrivate
{
    DiyaObject  parent;
    DiyaWayland * wayland;
    GtkApplication* app;
    DiyaVirtualKeyboard* vkb;
    gchar* name;
} DiyaShellPrivate;

G_DEFINE_TYPE_WITH_PRIVATE(DiyaShell, diya_shell, DIYA_TYPE_OBJECT);



static void diya_shell_dispose(GObject* object)
{
    g_debug("diya_shell_dispose: %s", diya_object_to_string(object));
    DiyaShell * self = DIYA_SHELL(object);
    DiyaShellPrivate* priv = diya_shell_get_instance_private(self);
    if(priv->wayland)
    {
        g_object_unref(priv->wayland);
    }
    if(priv->name)
    {
        g_free(priv->name);
        priv->name = NULL;
    }
    G_OBJECT_CLASS(diya_shell_parent_class)->dispose(object);
    if (priv->app)
    {
        g_application_quit(G_APPLICATION(priv->app));
    }
}

static void on_gtk_app_startup(GtkApplication *app, void *data)
{
    (void)app;
    DiyaShell *shell = data;
    DiyaShellPrivate* priv = diya_shell_get_instance_private(shell);
    priv->wayland =diya_wayland_new(DIYA_SHELL(shell));
    DiyaShellClass* class = DIYA_SHELL_GET_CLASS(shell);
    if(class->startup_handle)
    {
        class->startup_handle(shell);
    }
    if(class->monitor_changed_handle)
    {
        GListModel* monitor_list = gdk_display_get_monitors(gdk_display_get_default());
        g_debug("listen to monitor changed");
        g_signal_connect(monitor_list,"items-changed", G_CALLBACK(class->monitor_changed_handle), shell);
    }
}

static void on_gtk_app_active(GtkApplication *app, void *data)
{
    (void)app;
    DiyaShell *shell = data;
    DiyaShellClass* class = DIYA_SHELL_GET_CLASS(shell);
    if(class->active_handle)
    {
        class->active_handle(shell);
    }
}

static gboolean diya_shell_sigint_handle(gpointer data)
{
    g_object_unref(DIYA_SHELL(data));
    return true;
}

static void init_gtk_application(DiyaShell* self)
{
    DiyaShellPrivate* priv = diya_shell_get_instance_private(self);
    if(priv->app)
    {
        return;
    }
    priv->app = gtk_application_new(priv->name, G_APPLICATION_DEFAULT_FLAGS);
    g_signal_connect(priv->app, "startup", G_CALLBACK(on_gtk_app_startup), (void *)self);
    g_signal_connect(priv->app, "activate", G_CALLBACK(on_gtk_app_active), (void *)self);
    g_unix_signal_add(SIGINT,(GSourceFunc)diya_shell_sigint_handle,(void*)self);
   // g_signal_connect(priv->app, "shutdown", G_CALLBACK(shutdown), (void*)shell);
}

static void diya_shell_set_property(GObject *object, guint property_id, const GValue *value, GParamSpec *pspec)
{
    DiyaShell * self = DIYA_SHELL(object);
    DiyaShellPrivate* priv = diya_shell_get_instance_private(self);

    switch (property_id)
    {
    case PROP_SHELL_NAME:
        if(priv->name)
        {
            return;
        }
        priv->name = g_strdup(g_value_get_string(value));
        init_gtk_application(self);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}

static void diya_shell_get_property(GObject *object, guint property_id, GValue *value, GParamSpec *pspec)
{
    DiyaShell * self = DIYA_SHELL(object);
    DiyaShellPrivate* priv = diya_shell_get_instance_private(self);
    switch (property_id)
    {
    case PROP_SHELL_WAYLAND:
        g_value_set_pointer(value, priv->wayland);
        break;
    case PROP_SHELL_APP:
        g_value_set_pointer(value, priv->app);
        break;
    case PROP_SHELL_NAME:
        g_value_set_string(value, priv->name);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}


static void diya_shell_init(DiyaShell *self)
{
    DiyaShellPrivate* priv = diya_shell_get_instance_private(self);
    priv->wayland = NULL;
    priv->app = NULL;
    priv->name = NULL;
}

DiyaWayland * diya_shell_get_wayland(DiyaShell* shell)
{
    DiyaShellPrivate* priv = diya_shell_get_instance_private(shell);
    assert(DIYA_IS_WAYLAND(priv->wayland));
    return priv->wayland;
}

GtkApplication* diya_shell_get_application(DiyaShell* shell)
{
    DiyaShellPrivate* priv = diya_shell_get_instance_private(shell);
    assert(GTK_IS_APPLICATION(priv->app));
    return priv->app;
}

const char* diya_shell_get_name(DiyaShell* shell)
{
    DiyaShellPrivate* priv = diya_shell_get_instance_private(shell);
    return priv->name;
}

static const gchar *diya_shell_to_string(DiyaObject *object)
{
    return diya_shell_get_name(DIYA_SHELL(object));
}

static void diya_shell_class_init(DiyaShellClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(class);
    gobject_class->dispose = diya_shell_dispose;
    gobject_class->set_property = diya_shell_set_property;
    gobject_class->get_property = diya_shell_get_property;

    DiyaObjectClass *base_class = DIYA_OBJECT_CLASS(class);
    base_class->to_string = diya_shell_to_string;

    class->foreign_register = NULL;
    class->virtual_keyboard_register = diya_virtual_keyboard_register;
    class->monitor_changed_handle = NULL;
    class->startup_handle = NULL;
    class->active_handle = NULL;
 
    shell_properties[PROP_SHELL_WAYLAND] = g_param_spec_pointer("wayland", NULL, "Shell wayland", G_PARAM_READABLE); // 
    shell_properties[PROP_SHELL_NAME] = g_param_spec_string("name", NULL, "Shell name", NULL, G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY);
    shell_properties[PROP_SHELL_APP] = g_param_spec_pointer("application", NULL, "Shell application", G_PARAM_READABLE);

    g_object_class_install_properties (gobject_class, N_PROPERTIES, shell_properties);
}

int diya_shell_run(DiyaShell* shell, int argc, char **argv)
{   
    DiyaShellPrivate* priv = diya_shell_get_instance_private(shell);
    assert(GTK_IS_APPLICATION(priv->app));
    int status = g_application_run(G_APPLICATION(priv->app),argc,argv);
    g_object_unref(priv->app);
    return status;
}